#include <LiquidCrystal_I2C.h>
#include <EEPROM.h>

// melody for initialization
#define NOTE_B0  31
#define NOTE_C1  33
#define NOTE_CS1 35
#define NOTE_D1  37
#define NOTE_DS1 39
#define NOTE_E1  41
#define NOTE_F1  44
#define NOTE_FS1 46
#define NOTE_G1  49
#define NOTE_GS1 52
#define NOTE_A1  55
#define NOTE_AS1 58
#define NOTE_B1  62
#define NOTE_C2  65
#define NOTE_CS2 69
#define NOTE_D2  73
#define NOTE_DS2 78
#define NOTE_E2  82
#define NOTE_F2  87
#define NOTE_FS2 93
#define NOTE_G2  98
#define NOTE_GS2 104
#define NOTE_A2  110
#define NOTE_AS2 117
#define NOTE_B2  123
#define NOTE_C3  131
#define NOTE_CS3 139
#define NOTE_D3  147
#define NOTE_DS3 156
#define NOTE_E3  165
#define NOTE_F3  175
#define NOTE_FS3 185
#define NOTE_G3  196
#define NOTE_GS3 208
#define NOTE_A3  220
#define NOTE_AS3 233
#define NOTE_B3  247
#define NOTE_C4  262
#define NOTE_CS4 277
#define NOTE_D4  294
#define NOTE_DS4 311
#define NOTE_E4  330
#define NOTE_F4  349
#define NOTE_FS4 370
#define NOTE_G4  392
#define NOTE_GS4 415
#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  494
#define NOTE_C5  523
#define NOTE_CS5 554
#define NOTE_D5  587
#define NOTE_DS5 622
#define NOTE_E5  659
#define NOTE_F5  698
#define NOTE_FS5 740
#define NOTE_G5  784
#define NOTE_GS5 831
#define NOTE_A5  880
#define NOTE_AS5 932
#define NOTE_B5  988
#define NOTE_C6  1047
#define NOTE_CS6 1109
#define NOTE_D6  1175
#define NOTE_DS6 1245
#define NOTE_E6  1319
#define NOTE_F6  1397
#define NOTE_FS6 1480
#define NOTE_G6  1568
#define NOTE_GS6 1661
#define NOTE_A6  1760
#define NOTE_AS6 1865
#define NOTE_B6  1976
#define NOTE_C7  2093
#define NOTE_CS7 2217
#define NOTE_D7  2349
#define NOTE_DS7 2489
#define NOTE_E7  2637
#define NOTE_F7  2794
#define NOTE_FS7 2960
#define NOTE_G7  3136
#define NOTE_GS7 3322
#define NOTE_A7  3520
#define NOTE_AS7 3729
#define NOTE_B7  3951
#define NOTE_C8  4186
#define NOTE_CS8 4435
#define NOTE_D8  4699
#define NOTE_DS8 4978
#define REST      0

//buzzer to arduino pin 9
const int buzzer = 9;

//melody for game over
int melody[] = {
  NOTE_B4, 16, NOTE_B5, 16, NOTE_FS5, 16, NOTE_DS5, 16, //1
  NOTE_B5, 32, NOTE_FS5, -16, NOTE_DS5, 8, NOTE_C5, 16,
  NOTE_C6, 16, NOTE_G6, 16, NOTE_E6, 16, NOTE_C6, 32, NOTE_G6, -16, NOTE_E6, 8,

  NOTE_B4, 16,  NOTE_B5, 16,  NOTE_FS5, 16,   NOTE_DS5, 16,  NOTE_B5, 32,  //2
  NOTE_FS5, -16, NOTE_DS5, 8,  NOTE_DS5, 32, NOTE_E5, 32,  NOTE_F5, 32,
  NOTE_F5, 32,  NOTE_FS5, 32,  NOTE_G5, 32,  NOTE_G5, 32, NOTE_GS5, 32,  NOTE_A5, 16, NOTE_B5, 8
};

// sizeof gives the number of bytes, each int value is composed of two bytes (16 bits)
// there are two values per note (pitch and duration), so for each note there are four bytes
int notes = sizeof(melody) / sizeof(melody[0]) / 2;

int tempo = 105;

// this calculates the duration of a whole note in ms
int wholenote = (60000 * 4) / tempo;

int divider = 0, noteDuration = 0;
// I2C address 0x27, 16 column and 2 rows
LiquidCrystal_I2C lcd(0x27, 16, 2); 

uint16_t  xValue, yValue;

//states for pacman
static const int SPRITE_PACMAN_OPEN = 0;
static const int SPRITE_PACMAN_CLOSED = 1;
static const int SPRITE_GHOST = 2;
static const int SPRITE_HEART = 3;
static const int SPRITE_SMILEY = 4;

//types of sprites
static const int TYPE_NONE = 0;
static const int TYPE_HEART = 1;
static const int TYPE_GHOST = 2;
static const int MAX_SPRITES = 10;

// game states
static const int STATE_INTRO = 0;
static const int STATE_PLAY = 1;
static const int STATE_GAMEOVER = 2;

static const int HIGHSCORE_ADDRESS = 0;
static const int TOP = 0;
static const int BOTTOM = 1;

//struct for a sprite(position and type)
struct sprite {
  int x;
  int y;
  int type;
};

volatile bool joystickMoved = false;

struct sprite sprites[MAX_SPRITES];

long timeToMove, timeToAnimate, timeToDebounce;
int state, score, highScore, gameSpeed, pacmanX, pacmanY, ghostOdds;
boolean mouthState, smile;
static const byte spriteBitmaps[8][8] = {
  {0x7, 0xF, 0x1E, 0x1C, 0x1C, 0x1E, 0xF, 0x7},
  {0x0, 0xF, 0x1F, 0x1F, 0x1E, 0x1F, 0xF, 0x0},
  {0x19, 0x1F, 0x15, 0x1F, 0x11, 0x1F, 0x1D, 0xC},
  {0x0, 0x0, 0x0, 0xA, 0x15, 0x11, 0xA, 0x4},
  {0x0, 0xA, 0x0, 0x0, 0x11, 0xE, 0x0, 0x0},
  {0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x0, 0x0, 0x0},
  {0x0, 0x0, 0x0, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F},
  {0x1F, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0}
};

// current row of pacman
int currentRow = 1;
int previousRow = 1;

void playStartSong() {
  // Remember, the array is twice the number of notes (notes + durations)
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {

    // calculates the duration of each note
    divider = melody[thisNote + 1];
    if (divider > 0) {
      // regular note, just proceed
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      // dotted notes are represented with negative durations!!
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }

    // we only play the note for 90% of the duration, leaving 10% as a pause
    tone(buzzer, melody[thisNote], noteDuration * 0.9);

    // Wait for the specief duration before playing the next note.
    delay(noteDuration);

    // stop the waveform generation before the next note.
    noTone(buzzer);
  }
}

void playGameOverSound() {
    int melody[] = {262, 196,196, 220, 196,0, 247, 262};

    // note durations: 4 = quarter note, 8 = eighth note, etc.:
    int noteDurations[] = {4, 8, 8, 4,4,4,4,4 };

    for (int thisNote = 0; thisNote < 8; thisNote++) {
        int noteDuration = 1000/noteDurations[thisNote];
        tone(buzzer, melody[thisNote],noteDuration);
        int pauseBetweenNotes = noteDuration * 1.30;
        delay(pauseBetweenNotes);
        noTone(buzzer);
    }
}

// Funcție pentru tratarea întreruperii
ISR(PCINT1_vect) {
  if (!(PINC & (1 << PC1))) {
    joystickMoved = true;
  }
}

void setup() {
  // initializes the LCD module
  lcd.init();

  cli();  // Disable global interrupts

  // Configure interrupt for pin A0
  PCICR |= (1 << PCIE1);    // Enable interrupts for PCINT8-14 group (A0-A6)
  PCMSK1 |= (1 << PC1);    // Enable interrupt for pin A0

  sei();  // Enable global interrupts

  // initializes the serial communication at a baud rate of 9600.
  Serial.begin(9600);

  // create custom characters on the LCD
  for (int i = 0; i < (sizeof(spriteBitmaps) / 8); i++) {
    lcd.createChar(i, spriteBitmaps[i]);
  }
  delay(500);

  // initializes the LCD module to operate in 16x2 mode
  lcd.begin(16, 2);

  // sets the seed for the random number generator
  randomSeed(analogRead(1));

  // initialize game and variables
  state = STATE_INTRO;
  initVars();

  // Set buzzer - pin 9 as an output  
  DDRB |= (1 << DDB1);

  // Set A0 and A1 as input pins
  DDRC &= ~(1 << PC0); // A0
  DDRC &= ~(1 << PC1); // A1

  ADMUX = 0;
  /* ADC1 - channel 1 */
  ADMUX |= (1 << MUX0);
  /* AVCC with external capacitor at AREF pin */
  ADMUX |= (1 << REFS0);
  ADCSRA = 0;
  /* set prescaler at 128 */
  ADCSRA |= (7 << ADPS0);
  /* enable ADC */
  ADCSRA |= (1 << ADEN);
}
 
void loop() {
  ADMUX |= (1 << MUX0);
  /* start conversion */
  ADCSRA |= (1 << ADSC);
  /* wait until conversion is complete */
  while (!(ADCSRA & (1 << ADIF)));
  yValue = ADC;
 
  cli();  // Disable global interrupts to prevent changes during checks

  if (joystickMoved) {
    // Corresponding action when the joystick is moved
    Serial.println("Interrupt generated: Joystick moved up!");
    joystickMoved = false;
  }

  sei();  // Enable global interrupts again

  switch (state) {
    case STATE_INTRO: {
        //start game song
         playStartSong(); 
        //prepare for game
         intro(); 
        break; 
    }

    case STATE_PLAY: {
      previousRow = currentRow;
      // check for joystick position to move pacman
      if (yValue < 100) {
        currentRow = 0;
      } else if (yValue > 900) {
        currentRow = 1;
      }
      play(); 
      break;
    } 

    case STATE_GAMEOVER: { 
       //game over sound
        playGameOverSound(); 
        //show game over message
        gameover(); 
        break;
    }
  }
}

//initialize variables for a new game
void initVars() {
  for (int i = 0; i < MAX_SPRITES; i++) {
    sprites[i] = {0, 0, TYPE_NONE};
  }
  highScore = EEPROM.read(HIGHSCORE_ADDRESS);
  timeToMove = 0;
  timeToAnimate = 0;
  timeToDebounce = 0;
  score = 0;
  gameSpeed = 600;
  ghostOdds = 6;
  pacmanX = 1;
  pacmanY = 0;
  mouthState = false;
  smile = false;
}

// show intro messages
void intro() {
  lcd.clear();
  lcd.setCursor(3, 0); 
  lcd.print("WELCOME TO"); 
  lcd.setCursor(1, 1); 
  lcd.print("SABINA'S GAME");
  delay(2000);

  lcd.clear(); 
  lcd.setCursor(3, 0); 
  lcd.print("COLLECT: ");
  lcd.write(byte(SPRITE_HEART));
  lcd.setCursor(4, 1);
  lcd.print("AVOID: "); 
  lcd.write(byte(SPRITE_GHOST));
  delay(4000);

  lcd.clear(); 
  lcd.setCursor(1, 0); 
  lcd.print("ARE YOU READY?"); 
  lcd.setCursor(6, 1); 
  lcd.print("GO!");
  delay(2000);

  //show animation and start game 
  animation(1);
  state = STATE_PLAY;
}

void play() {
  // draw all the sprites on the lcd
  drawScreen();

  long now = millis();
  if (now > timeToMove) {
    //move all sprites to generate the ilusion of a game
    moveLeft();
    
    // generate random ghosts at random moments
    if (!random(0, ghostOdds)) {
      spawn(TYPE_GHOST);
    }

    // generate random hearts at random moments
    if (!random(0, 3)) {
      spawn(TYPE_HEART);
    }
    timeToMove = now + gameSpeed;
  }

  //check for any collision
  int c = at(pacmanX, pacmanY);
  if (c == TYPE_HEART) {
    eatHeart();
    // pacman sprite successfully eats a heart sprite and the score must be updated
    score++;

  //when there is a collision with a ghost, game over
  } else if (c == TYPE_GHOST) {
    state = STATE_GAMEOVER;
  }
}

void gameover() {
  //show game over message and animation
  animation(0);
  lcd.setCursor(3, 0); 
  lcd.print("GAME OVER"); 

  //show score for current game
  lcd.clear(); 
  lcd.setCursor(2, 0); 
  lcd.print("YOUR SCORE:"); 
  lcd.setCursor(7, 1);
  lcd.print(score);
  delay(2000);

  //check for new highScore and update it
  if (score > highScore) {
    lcd.clear();
    lcd.setCursor(1, 0); 
    lcd.print("NEW HIGHSCORE!");
    EEPROM.write(HIGHSCORE_ADDRESS, score);
    highScore = score;
    delay(2000);
  }

  lcd.clear();
  lcd.setCursor(3, 0); 
  lcd.print("TRY AGAIN"); 
  lcd.setCursor(2, 1); 
  lcd.print("Highscore: ");
  lcd.print(highScore);
  delay(3000);

  //reinitialize game
  initVars();
  state = STATE_PLAY;
  lcd.clear();
}

void drawScreen() {
  //draw figures on the screen
  for (int i = 0; i < MAX_SPRITES; i++) {
    drawSprite(i);
  }

  //check if pacman vertical position has updated from prev frame
  // and update it
  if (currentRow != previousRow) {
    hidePacman();
    pacmanY = currentRow;
  } 
  drawPacman();
}

// when pacman encounters a heart sprite
void eatHeart() {
  //checks each sprite in the game to find a heart sprite 
  //that overlaps with the Pacman's position
  for (int i = 0; i < MAX_SPRITES; i++) {
    if (sprites[i].x == pacmanX && sprites[i].y == pacmanY && sprites[i].type == TYPE_HEART) {
      //pacman is drawn a a smile figure and the heart sprite is deleted
      smile = true;
      deleteSprite(i);
      return;
    }
  }
}

// spawning new sprites in the game (hearts/ghosts)
void spawn(int type) {
  //generate random position
  int x = 15;
  int y = random(0, 2);
  for (int i = 0; i < MAX_SPRITES; i++) {
    int spriteType = sprites[i].type;
    //check if position is free
    if (spriteType == TYPE_NONE) {
      if (((type == TYPE_GHOST) && okayToSpawnGhost(y)) || 
      ((type == TYPE_HEART) && okayToSpawnHeart(y))) {
        //create new figure
        createSprite(i, x, y, type);
      }
      return;
    }
  }
}

// // creating a sprite and assigning its properties.
void createSprite(int s, int x, int y, int type) {
  sprites[s].x = x;
  sprites[s].y = y;
  sprites[s].type = type;
}

// checks if there is any sprite present at the specified coordinates
int at(int x, int y) {
  for (int i = 0; i < MAX_SPRITES; i++) {
    if (sprites[i].x == x && sprites[i].y == y && sprites[i].type != TYPE_NONE) {
      //if sprite is presenet, return its type
      return sprites[i].type;
    }
  }
  //else, there isn't any sprite at this position
  return TYPE_NONE;
}

// checks if it is okay to spawn a ghost sprite at the specified vertical position
int okayToSpawnGhost(int pos) {
  // check if there is any sprite already present at the coordinates
  if (at(15, pos) != TYPE_NONE) {
    return 0;
  } else if (at(15, !pos) == TYPE_GHOST) {
    return 0;
  //it shown exist a space for pacman to be able to avoid ghosts
  } else if ((pos == TOP) && (at(14, BOTTOM) == TYPE_GHOST)) {
    return 0;
  } else if ((pos == BOTTOM) && (at(14, TOP) == TYPE_GHOST)) {
    return 0;
  }
  return 1;
}

// checks if it is okay to spawn a heart sprite at the specified vertical position 
int okayToSpawnHeart(int pos) {
  if (at(15, pos) != TYPE_NONE) {
    return 0;
  }
  return 1;
}

// moving all the sprites in the game one position to the left
void moveLeft() {
  for (int i = 0; i < MAX_SPRITES; i++) {
    // if there is a sprite at this position, move it to the left
    if (sprites[i].type != TYPE_NONE) {
      // decrement the horizontal position and the vertical remains the same
      int x = sprites[i].x - 1;
      int y = sprites[i].y;
      moveSprite(i, x, y);
    } 
  }
}

// responsible for moving a sprite to a new position
// used to move to left all sprites
void moveSprite(int s, int x, int y) {
  if ((x < 0) || (x > 15)) {
    deleteSprite(s);
  } else {
    hideSprite(s);
    sprites[s].x = x;
    sprites[s].y = y;
  }
}

// draw a sprite on the LCD screen.
void drawSprite(int s) {
  int spriteType = sprites[s].type;
  if (spriteType == TYPE_NONE) {
    return;
  }

  //get sprite position to draw it
  int x = sprites[s].x;
  int y = sprites[s].y;
  if (!((x == pacmanX) && (y == pacmanY))) {
    lcd.setCursor(x, y);
    switch (spriteType) {
      case TYPE_GHOST: lcd.write(byte(SPRITE_GHOST)); break;
      case TYPE_HEART: lcd.write(byte(SPRITE_HEART)); break;
      default: lcd.write(byte(32)); break;
    }
  }
}

// hide a sprite from the LCD screen.
void hideSprite(int s) {
  lcd.setCursor(sprites[s].x, sprites[s].y);
  //writes a space character to hide sprite
  lcd.write(byte(32));
}

// draw the Pacman sprite on the LCD screen
void drawPacman() {
  if ((millis() > timeToAnimate) || (smile)) {
    // set the duration of the animation
    int wait = 350;
    lcd.setCursor(pacmanX, pacmanY);
    // wait for animation 
    if (smile) {
      lcd.write(byte(SPRITE_SMILEY));
      wait = 600;
      smile = false;

    // toggles to alternate between the open and closed mouth states
    } else if (mouthState) {
      lcd.write(byte(SPRITE_PACMAN_OPEN));
    } else {
      lcd.write(byte(SPRITE_PACMAN_CLOSED));
    }

  // updates the variable for the next animation frame.
    mouthState = !mouthState;
    timeToAnimate = millis() + wait;
  }
}

void hidePacman() {
  lcd.setCursor(pacmanX, pacmanY);
  lcd.write(byte(32));
}

//hide sprite and mark it as empty
void deleteSprite(int s) {
  hideSprite(s);
  sprites[s].x = 0;
  sprites[s].y = 0;
  sprites[s].type = TYPE_NONE;
}

// show animation for starting/finishing the game
void animation(int direction) {
  //  store the binary patterns for the animation frames
  byte animationOpen[6] = {255, 255, 5, 6, 7, 95};
  byte animationClose[6] = {7, 95, 5, 6, 255, 255};
  byte animationChars[6];

  // check animation type: up or down
  if (direction) {
    memcpy(animationChars, animationOpen, sizeof animationChars);
  } else {
    memcpy(animationChars, animationClose, sizeof animationChars);
  }

  // show animation on lcd
  lcd.clear();
  for (int frame = 0; frame < 3; frame++) {
    lcd.setCursor(0, 0);
    for (int i = 0; i < 16; i++) {
        lcd.write(animationChars[frame * 2]);
    }
    lcd.setCursor(0, 1);
    for (int i = 0; i < 16; i++) {
        lcd.write(animationChars[(frame * 2) + 1]);
    }
    delay(300);
    lcd.clear();
  }
}